use crate::api;
use crate::display::create_display_thread;
use crate::dither::{DitherMethod, DitheredImage};
use crate::eink::Palette;
use axum::extract::{Path, State};
use axum::http::{header, StatusCode};
use axum::response::{IntoResponse, Response};
use axum::routing::{get, post};
use axum::Router;
use base64::{engine::general_purpose::STANDARD, Engine as _};
use image::imageops::{resize, FilterType};
use include_dir::{include_dir, Dir, DirEntry};
use minijinja::{context, Environment};
use mime_guess::Mime;
use std::io::Cursor;
use std::sync::mpsc;
use std::sync::Arc;
use std::thread::JoinHandle;
use strum::VariantNames;
use tracing::{debug, warn};
use tracing::{error, info, instrument};

#[derive(thiserror::Error, Debug)]
pub enum ApiError {
    #[error("missing image field")]
    MissingImage,
}

#[derive(Clone)]
pub struct AppState {
    pub display_channel: mpsc::Sender<Box<DitheredImage>>,
    pub display_task: Arc<JoinHandle<()>>,
    pub templates: Arc<Environment<'static>>,
}

impl Default for AppState {
    #[must_use]
    fn default() -> Self {
        let (handle, tx) = create_display_thread();
        Self {
            display_channel: tx,
            display_task: Arc::new(handle),
            templates: Arc::new(make_environment().expect("valid")),
        }
    }
}

/// Represents an error that can be returned to the requester by implementing ``IntoResponse``.
/// This wraps ``anyhow::Error`` so that it can accept any error object.
pub struct AppError(anyhow::Error);

// Tell axum how to convert `AppError` into a response.
impl IntoResponse for AppError {
    fn into_response(self) -> Response {
        (
            StatusCode::INTERNAL_SERVER_ERROR,
            format!("Something went wrong: {}", self.0),
        )
            .into_response()
    }
}

// This enables using `?` on functions that return `Result<_, anyhow::Error>` to turn them into
// `Result<_, AppError>`. That way you don't need to do that manually.
impl<E> From<E> for AppError
where
    E: Into<anyhow::Error>,
{
    fn from(err: E) -> Self {
        Self(err.into())
    }
}

static TEMPLATES_DIR: Dir = include_dir!("$CARGO_MANIFEST_DIR/templates");
static ASSETS_DIR: Dir = include_dir!("$CARGO_MANIFEST_DIR/static");

#[instrument]
fn make_environment() -> Result<Environment<'static>, anyhow::Error> {
    let mut env = Environment::new();
    let mut entries = vec![&TEMPLATES_DIR];

    while let Some(dir) = entries.pop() {
        for entry in dir.entries() {
            match entry {
                DirEntry::Dir(d) => {
                    entries.push(d);
                }
                DirEntry::File(f) => {
                    debug!("adding {:?} to minijinja environment", f.path());
                    env.add_template(
                        f.path().to_str().expect("valid pathname"),
                        f.contents_utf8().expect("contents are valid"),
                    )?;
                }
            }
        }
    }

    info!(
        "loaded {} templates into environment",
        env.templates().count()
    );

    Ok(env)
}

pub fn make_app_router() -> Router {
    Router::new()
        .route("/app", get(app_handler))
        .route("/app/preview", post(app_preview))
        .nest("/api", api::router())
        .route("/static/*path", get(asset_handler))
        .with_state(AppState::default())
}

#[instrument(skip(ctx))]
async fn app_handler(State(ctx): State<AppState>) -> Result<impl IntoResponse, AppError> {
    let template = ctx.templates.get_template("app.html")?;

    let content = template.render(context! {
        palettes => Palette::VARIANTS,
        dither_methods => DitherMethod::VARIANTS,
        version => env!("CARGO_PKG_VERSION"),
    })?;

    let headers = [(header::CONTENT_TYPE, mime::TEXT_HTML_UTF_8.to_string())];

    Ok((StatusCode::OK, headers, content))
}

async fn app_preview(
    State(ctx): State<AppState>,
    img_req: api::ImageRequest,
) -> Result<impl IntoResponse, AppError> {
    let mut buf = DitheredImage::new(800, 480, img_req.palette.value().to_vec());
    let resized = resize(&*img_req.image, 800, 480, FilterType::Lanczos3);
    {
        let mut dither = img_req.dither_method.get_ditherer();
        dither.dither(&resized, &mut buf);
    }
    // Convert buf into a png image.
    let img = buf.into_rgbimage();

    let mut buffer = Cursor::new(Vec::new());
    img.write_to(&mut buffer, image::ImageFormat::Png)?;

    let template = ctx.templates.get_template("preview.html")?;

    let datavec = STANDARD.encode(buffer.into_inner());
    let content = template.render(context! {
        image_contents => datavec
    })?;
    let headers = [(header::CONTENT_TYPE, mime::TEXT_HTML_UTF_8.to_string())];
    Ok((StatusCode::OK, headers, content))
}

async fn asset_handler(Path(s): Path<String>) -> Result<impl IntoResponse, AppError> {
    let Some(file) = ASSETS_DIR.get_file(&s) else {
        panic!("fixme");
    };

    let mime = mime_guess::from_path(s).first_or_text_plain();

    let headers = [
        (header::CONTENT_TYPE, mime.essence_str().to_string())
    ];

    Ok((StatusCode::OK, headers, file.contents()))
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_environment() -> anyhow::Result<()> {
        make_environment()?;
        Ok(())
    }
}
